import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.HashMap;
import java.util.Properties;
import java.util.TimeZone;
import java.util.Map.Entry;
import java.util.logging.LogManager;

import javax.mail.MessagingException;
import javax.mail.internet.AddressException;

import com.gargoylesoftware.htmlunit.ThreadedRefreshHandler;
import com.gargoylesoftware.htmlunit.WebClient;
import com.gargoylesoftware.htmlunit.html.HtmlForm;
import com.gargoylesoftware.htmlunit.html.HtmlPage;
import com.gargoylesoftware.htmlunit.html.HtmlTable;
import com.gargoylesoftware.htmlunit.html.HtmlTableCell;
import com.gargoylesoftware.htmlunit.html.HtmlTableRow;

public class InterviewChecker {
  private static final SimpleDateFormat df = new SimpleDateFormat("H:m");
  private static final String PROPERTIES_FILENAME = "properties.txt";
  private static final String JOBMINE_USER_KEY = "JobmineUserID";
  private static final String JOBMINE_PASS_KEY = "JobminePassword";
  private static final String GMAIL_USER_KEY = "GmailAddress";
  private static final String GMAIL_PASS_KEY = "GmailPassword";
  private static final String NOTIFY_EMAIL_KEY = "NotifyEmailAddress";
  private static final String helpOpt = "-h";
  private static final String propertiesFileOpt = "-f";
  private static final String jobmineUserOpt = "-ju";
  private static final String jobminePassOpt = "-jp";
  private static final String gmailAddressOpt = "-ga";
  private static final String gmailPassOpt = "-gp";
  private static final String notifyAddressOpt = "-n";
  static {
    df.setTimeZone(TimeZone.getTimeZone("America/Toronto"));
  }
  private final long lastModified;
  private boolean validNotifier;
  private boolean sendTest;
  private boolean firstCheck;
  private String jobmineUser;
  private String jobminePass;
  private Notifier notifier;
  private String[][] t = null;
  private HtmlPage resp = null;
  private File objectFile;

  private InterviewChecker(String propertiesFilename, Properties override)
  throws IOException, AddressException, ClassNotFoundException {
    sendTest = false;
    validNotifier = false;
    firstCheck = true;
    File propertiesFile = new File(propertiesFilename);
    this.lastModified = propertiesFile.lastModified();
    Properties props = getProps(propertiesFile);
    if (!override.isEmpty()) {
      if (override.containsKey(GMAIL_USER_KEY)
          || override.containsKey(GMAIL_PASS_KEY)
          || override.containsKey(NOTIFY_EMAIL_KEY)) sendTest = true;
      props.putAll(override);
    }
    readProps(props);
    init(props);
  }

  private void init(Properties props) throws IOException, AddressException, ClassNotFoundException {
    jobmineUser = getProp(props, JOBMINE_USER_KEY);
    jobminePass = getProp(props, JOBMINE_PASS_KEY);
    String gmailUser = getProp(props, GMAIL_USER_KEY);
    String gmailPass = getProp(props, GMAIL_PASS_KEY);
    String notifyEmail = getProp(props, NOTIFY_EMAIL_KEY);
    notifier = new Notifier(gmailUser, gmailPass, notifyEmail);

    objectFile = new File("." + jobmineUser);
    if (objectFile.exists()) {
      FileInputStream in = new FileInputStream(objectFile);
      ObjectInputStream ois = new ObjectInputStream(in);
      try {
        long readModified = ois.readLong();
        validNotifier = ois.readBoolean();
        t = (String[][]) ois.readObject();
        if (lastModified != readModified) validNotifier = false;
        if (t != null) {
          System.out.println("Prior number of interviews: " + numberOfInterviews(t));
        }
      }
      catch (Exception e) {}
      ois.close();
    }
  }

  private static Properties getProps(File file) throws IOException {
    Properties props = new Properties();
    try {
      FileInputStream in = new FileInputStream(file);
      props.load(in);
    }
    catch (FileNotFoundException e) {}
    return props;
  }

  private String getProp(Properties props, String key) throws IOException {
    if (props.containsKey(key)) return props.getProperty(key);
    else throw new IOException("Missing '" + key + "' property.");
  }

  private void readProps(Properties props) {
    if (!props.containsKey(JOBMINE_USER_KEY) || props.get(JOBMINE_USER_KEY).equals("myWatIAMUserID")) {
      System.out.print("Enter Jobmine user: ");
      props.put(JOBMINE_USER_KEY, System.console().readLine());
      System.out.print("Enter password for Jobmine user " + props.getProperty(JOBMINE_USER_KEY) + ": ");
      props.put(JOBMINE_PASS_KEY, new String(System.console().readPassword()));
    }
    else if (!props.containsKey(JOBMINE_PASS_KEY)) {
      System.out.print("Enter password for Jobmine user " + props.getProperty(JOBMINE_USER_KEY) + ": ");
      props.put(JOBMINE_PASS_KEY, new String(System.console().readPassword()));
    }
    if (!props.containsKey(GMAIL_USER_KEY) || props.get(GMAIL_USER_KEY).equals("gAccount@gmail.com")) {
      System.out.print("Enter Gmail address: ");
      props.put(GMAIL_USER_KEY, System.console().readLine());
      System.out.print("Enter password for Gmail address " + props.getProperty(GMAIL_USER_KEY) + ": ");
      props.put(GMAIL_PASS_KEY, new String(System.console().readPassword()));
      sendTest = true;
    }
    else if (!props.containsKey(GMAIL_PASS_KEY)) {
      System.out.print("Enter password for Gmail address " + props.getProperty(GMAIL_USER_KEY) + ": ");
      props.put(GMAIL_PASS_KEY, new String(System.console().readPassword()));
      sendTest = true;
    }
    if (!props.containsKey(NOTIFY_EMAIL_KEY) || props.get(NOTIFY_EMAIL_KEY).equals("sendHere@example.com")) {
      System.out.print("Enter email address to notify: ");
      props.put(NOTIFY_EMAIL_KEY, System.console().readLine());
      sendTest = true;
    }
  }

  private static int numberOfInterviews(String[][] t) {
    return (t.length == 2 && t[1][1].equals(" ")) ? 0 : t.length - 1;
  }

  private boolean shouldNotify(String[][] s) {
    if (!t[1][1].equals(s[1][1])) return true;
    if (t.length != s.length) return true;
    for (int i = 1; i < t.length; i++) {
      if (t[i][7].length() != 1 && s[i][7].length() == 1) return true;
    }
    return false;
  }

  public void clear() {
    if (resp != null) {
      resp.cleanUp();
      resp = null;
    }
  }

  private void saveState() {
    try {
      FileOutputStream os = new FileOutputStream(objectFile);
      ObjectOutputStream oos = new ObjectOutputStream(os);
      oos.writeLong(lastModified);
      oos.writeBoolean(validNotifier);
      oos.writeObject(t);
      oos.flush();
      oos.close();
    }
    catch (IOException e) {
      System.out.println("Unable to save new state.");
      e.printStackTrace();
    }
  }

  public void check() {
    boolean success = resp != null;
    if (!success) {
      try {
        WebClient wc = new WebClient();
        wc.setCssEnabled(false);
        wc.setJavaScriptEnabled(false);
        wc.setRefreshHandler(new ThreadedRefreshHandler());
        resp = wc.getPage("https://jobmine.ccol.uwaterloo.ca/psp/SS/?cmd=login");
        HtmlForm form = resp.getFormByName("login");
        form.getInputByName("userid").setValueAttribute(jobmineUser);
        form.getInputByName("pwd").setValueAttribute(jobminePass);
        resp = form.getInputByName("submit").click();
        resp = resp.getAnchorByText("Interviews").click();
        success = true;
      }
      catch (Exception e) {
        System.out.println("Failed to access Main page. Will retry later.");
        System.out.println("Your credentials may be invalid. Verify properties.");
        e.printStackTrace();
        this.clear();
      }
    }
    if (success) {
      try {
        resp = resp.getAnchorByText("Interviews").click();
        HtmlTable table = (HtmlTable) ((HtmlPage) resp.getFrameByName("TargetContent").getEnclosedPage()).getAnchorByText("Job ID").getEnclosingElement(HtmlTable.TAG_NAME);
        String[][] s = new String[table.getRowCount()][table.getRow(0).getCells().size()];
        {
          int i = 0;
          for (HtmlTableRow row : table.getRows()) {
            int j = 0;
            for (HtmlTableCell cell : row.getCells()) {
              s[i][j] = cell.asText();
              j++;
            }
            i++;
          }
        }
//        resp = resp.getAnchorByText("Home").click();
        if (firstCheck) {
          System.out.print("Check successful. ");
        }
        if (t == null) {
          t = s;
          System.out.println("Initial number of interviews: " + numberOfInterviews(t));
          saveState();
        } else if (shouldNotify(s)) {
          int before = numberOfInterviews(t);
          int after = numberOfInterviews(s);
          System.out.println("Interviews changed from " + before + " to " + after + ".");
          try {
            notifier.send("Interview change occurred", "Check Jobmine. You may need to book a time slot.");
            t = s;
            saveState();
          }
          catch (MessagingException e) {
            System.out.println("Unable to send notification:");
            e.printStackTrace();
          }
        } else if (firstCheck) {
          System.out.println("No changes or no significant changes.");
        }
        if (firstCheck) {
          System.out.println("Will continue to check every 10 minutes, silently unless there is a change to report.");
          firstCheck = false;
        }
      }
      catch (Exception e) {
        System.out.println("Failed to read Interview count. Will retry later.");
        e.printStackTrace();
        this.clear();
      }
    }
  }

  private void testNotification() throws MessagingException {
    // only send test notification if necessary
    if (sendTest || !validNotifier) {
      System.out.println("Sending initial test notification...");
      notifier.send("Initial Test Notification", "This is a test.");
      if (!sendTest) validNotifier = true;
      saveState();
      System.out.println("Test notification was successful if you received an email.");
    }
  }

  private static int untilOpen(int d, int h, int m) {
    // check if before opening time.
    switch (d) {
    case Calendar.TUESDAY: case Calendar.WEDNESDAY:
    case Calendar.THURSDAY: case Calendar.FRIDAY:
    case Calendar.SATURDAY:
      int openHour = (d == Calendar.WEDNESDAY) ? 8 : 7;
      if (h <= openHour) {
        int r = 60*(openHour-h) + (2-m);
        return (r>0) ? r : 0;
      }
    }
    // check if after closing time.
    switch (d) {
    case Calendar.MONDAY:
    case Calendar.TUESDAY: case Calendar.WEDNESDAY:
    case Calendar.THURSDAY: case Calendar.FRIDAY:
      if (h == 23 && m > 57) {
        int openHour = (d == Calendar.TUESDAY) ? 8 : 7;
        return 60*openHour + (62-m);
      }
    }
    return 0;
  }

//  private static boolean isOpen(int d, int h, int m) {
//    if (h > (d == Calendar.WEDNESDAY?8:7) && h < 23) return true;
//    if (h == (d == Calendar.WEDNESDAY?8:7) && m >= 5) return true;
//    if (h == 23 && m <= 55) return true;
//    return false;
//  }

  private static void printHelp() {
    System.out.println("Usage: java -jar InterviewChecker.jar [ options ]");
    System.out.println();
    System.out.println("Options:");
    System.out.println("Command line options take precedence over properties in the properties file.");
    System.out.println("You will be prompted for missing information.");
    System.out.println();
    System.out.println("    " + helpOpt +           "               Display this message");
    System.out.println("    " + propertiesFileOpt + "  <file>       Defaults to " + PROPERTIES_FILENAME);
    System.out.println("    " + jobmineUserOpt +     " <user>       UWDir user");
    System.out.println("    " + jobminePassOpt +     " <pass>       UWDir password");
    System.out.println("    " + gmailAddressOpt +    " <address>    Gmail sender, example: gAccount@gmail.com");
    System.out.println("    " + gmailPassOpt +       " <pass>       Gmail password");
    System.out.println("    " + notifyAddressOpt +  "  <address>    Notify email, example: sendHere@example.com");
    System.out.flush();
  }

  public static void main(String[] args) {
    InterviewChecker checker = null;
    String propertiesFilename = PROPERTIES_FILENAME;
    HashMap<String,String> optProp = new HashMap<String,String>(5);
    optProp.put(jobmineUserOpt, JOBMINE_USER_KEY);
    optProp.put(jobminePassOpt, JOBMINE_PASS_KEY);
    optProp.put(gmailAddressOpt, GMAIL_USER_KEY);
    optProp.put(gmailPassOpt, GMAIL_PASS_KEY);
    optProp.put(notifyAddressOpt, NOTIFY_EMAIL_KEY);
    HashMap<String,String> optArgs = new HashMap<String,String>();
    optArgs.put(propertiesFileOpt, null);
    optArgs.put(jobmineUserOpt, null);
    optArgs.put(jobminePassOpt, null);
    optArgs.put(gmailAddressOpt, null);
    optArgs.put(gmailPassOpt, null);
    optArgs.put(notifyAddressOpt, null);

    try {
      Properties options = new Properties();
      for (int i = 0; i < args.length; i++) {
        if (args[i].equals(helpOpt)) {
          printHelp();
          System.exit(0);
        }
        else if (args[i].startsWith("-")) {
          if (optArgs.containsKey(args[i])) {
            if (!args[i+1].startsWith("-")) {
              if (optArgs.get(args[i]) == null) {
                optArgs.put(args[i], args[i+1]);
                i++;
              }
              else throw new IllegalArgumentException("Duplicate option '" + args[i] + "'");
            }
            else throw new IllegalArgumentException("Option '" + args[i] + "' requires an argument");
          }
          else throw new IllegalArgumentException("Illegal option: " + args[i]);
        }
        else throw new IllegalArgumentException("Illegal argument: " + args[i]);
      }
      if (optArgs.get(propertiesFileOpt) != null) propertiesFilename = optArgs.get(propertiesFileOpt);
      for (Entry<String,String> optArg : optArgs.entrySet()) {
        if (optArg.getValue() != null && optProp.get(optArg.getKey()) != null) {
          options.put(optProp.get(optArg.getKey()), optArg.getValue());
        }
      }
      checker = new InterviewChecker(propertiesFilename, options);
      LogManager.getLogManager().reset();
    }
    catch (IllegalArgumentException e) {
      System.err.println("ERROR: " + e.getMessage());
      printHelp();
      System.exit(1);
    }
    catch (AddressException e) {
      System.err.println("ERROR: Check email address format.");
      System.err.println(e);
      System.exit(1);
    }
    catch (Exception e) {
      System.err.println("ERROR: " + e);
      e.printStackTrace(System.err);
      System.exit(1);
    }

    try {checker.testNotification();}
    catch (MessagingException e) {
      System.err.println("ERROR: Test notification failed. Verify credentials.");
      System.exit(1);
    }

    boolean neverTried = true;
    Calendar c = df.getCalendar();
    while (true) {
      c.setTimeInMillis(System.currentTimeMillis());
      int d = c.get(Calendar.DAY_OF_WEEK);
      int h = c.get(Calendar.HOUR_OF_DAY);
      int m = c.get(Calendar.MINUTE);
      int t = untilOpen(d, h, m); // minutes until open
      if (t == 0) {
        if (neverTried) {
          System.out.println("Checking now...");
          neverTried = false;
        }
        checker.check();
        m = 0;
        try {Thread.sleep(60000 * 10);}
        catch (Exception e) {System.exit(0);}
      }
      else {
        if (neverTried) {
          System.out.println("Time in Waterloo is " + df.format(c.getTime()) + ".");
          System.out.println("JobMine is asleep. Will check when JobMine wakes up.");
        }
        checker.clear();
        try {Thread.sleep(60000 * t);}
        catch (Exception e) {System.exit(0);}
      }
    }
  }

}
